﻿using System;
using System.Linq;
using System.Reflection;
using Cfg.Attributes;
using Cfg.Exceptions;
using Cfg.Helpers;
using Cfg.ValidationAttributes;

namespace Cfg.ConfigValues
{
    public class ValueBuilder
    {
        private readonly Type _type;
        private readonly object[] _attributes;
        private readonly string _context;

        /// <summary>
        /// Original value in config file
        /// </summary>
        private readonly string _stringValue;

        /// <summary>
        /// Formatted value
        /// </summary>
        private readonly string _finalizedStringValue;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="type"></param>
        /// <param name="finalizedString"></param>
        public ValueBuilder(Type type, string finalizedString)
            : this(type, null, null, finalizedString, finalizedString)
        {}

        /// <summary>
        /// 
        /// </summary>
        /// <param name="type"></param>
        /// <param name="attributes"></param>
        /// <param name="context"></param>
        /// <param name="stringValue"></param>
        /// <param name="finalizedString"></param>
        public ValueBuilder(Type type, object[] attributes, string context, string stringValue, string finalizedString)
        {
            _type = type;
            _attributes = attributes ?? new object[0];
            _context = context;

            _stringValue = stringValue;
            _finalizedStringValue = finalizedString;
        }

        /// <summary>
        /// Get attribute
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        private T Attr<T>()
        {
            return _attributes.OfType<T>().FirstOrDefault();
        }

        /// <summary>
        /// Gets default value defined via DefaultAttribute
        /// </summary>
        /// <returns></returns>
        public object DefaultFromAttribute()
        {
            var defaultAttribute = Attr<DefaultAttribute>();
            if (defaultAttribute == null || (defaultAttribute.Contexts != null && defaultAttribute.Contexts.Length != 0 && !defaultAttribute.Contexts.Contains(_context)))
                return null;
            
            return defaultAttribute.DefaultValue;
        }

        /// <summary>
        /// Gets current types default value
        /// </summary>
        /// <returns></returns>
        public object Default()
        {
            if (_type.IsValueType)
            {
                return Activator.CreateInstance(_type);
            }
            return null;
        }

        /// <summary>
        /// Get parsed value object
        /// </summary>
        /// <returns></returns>
        public object GetValue()
        {
            // todo - check if _type is an array

            // return default value if config value not set
            if (string.IsNullOrEmpty(_finalizedStringValue))
            {
                // note - default value is NOT validated
                var defaultValue = DefaultFromAttribute();
                if (defaultValue != null)
                    return defaultValue;
            }

            // validate finalized string
            foreach (var attribute in _attributes.OfType<CfgFinalizedStringValueValidatorBaseAttribute>())
            {
                attribute.Validate(_finalizedStringValue, _context);
            }
            
            object returnValue;

            if (_type.IsInheritedFrom(typeof(ConfigValueBase)))
            {
                var cvb = Activator.CreateInstance(_type) as ConfigValueBase;
                cvb.ConfigValue = _stringValue;
                cvb.FinalizedStringValue = _finalizedStringValue;
                cvb.Attributes = _attributes;
                cvb.Parse();

                returnValue = cvb;
            }
            else if (AppSettings.Settings.Parsers.ContainsKey(_type))
            {
                returnValue = AppSettings.Settings.Parsers[_type].DynamicInvoke(_finalizedStringValue);
            }
            /*
            else if (_type.HasParseExtension())
            {
                returnValue = Activator.CreateInstance(_type);
                var mi = _type.GetMethod("Parse", new[] {_type, typeof (string)});
                mi.Invoke(returnValue, new object[] {FinalizedStringValue});
            }
            **/
            else if (string.IsNullOrEmpty(_finalizedStringValue))
            {
                returnValue = Default();
            }
            else
            {
                try
                {
                    returnValue = StringConverter.Convert(_finalizedStringValue, _type);
                }
                catch (TargetInvocationException)
                {
                    throw new InvalidFormatException(string.Format("Invalid format. {0} is not type of {1}", _finalizedStringValue, _type));
                }
            }

            foreach (var validator in _attributes.OfType<CfgValueValidatorBaseAttribute>())
            {
                validator.Validate(returnValue, _context);
            }

            return returnValue;
        }
    }
}